/********************** localdeliver.c **********************/

/******************************************************************************
 * Copyright (c) 1995,1996 Donated to the public domain                       *
 *                                                                            *
 * Original author and maintainer: Marc Chatel   (mchatel@dial.oleane.com)    *
 * Last known maintainer:          Marc Chatel   (mchatel@dial.oleane.com)    *
 *                                                                            *
 * This file was created as part of the S4 (Secure System Setup Script) kit.  *
 * Permission is granted to any person or entity to do any of the following:  *
 *       a) use this file alone or in some other software                     *
 *       b) modify this file or include parts of this file in other files     *
 *       c) re-distribute this file AS IS or modified, for non-commercial     *
 *          or commercial purposes, alone or as part of some software package *
 *                                                                            *
 * No warranties of any kind, express or implied, on the functionality and    *
 * safety of the contents of this file. Use at your own risk!                 *
 *                                                                            *
 * If you do useful changes to this file (bug fixes, portability fixes,       *
 * enhancements), you should TRY to contact the current maintainer, who may   *
 * be maintaining a "latest greatest" version of the file. You do not HAVE    *
 * TO, but you should TRY. Promote software reuse! It helps everybody,        *
 * including you!                                                             *
 ******************************************************************************/

/* localdeliver
   ------------
   This is a "glue" program. It is only useful in very precise conditions...

   The following assumptions are made:

   a) sendmail is running non-privileged and chrooted, but you still need
         to handle mail delivery to local system users.

   b) We want to separate sendmail processing (the mail queue, sendmail config,
      and so on) from the "local" mailboxes. In other words, we would like to
      make it difficult for sendmail to look into the user mailboxes, and we
      would like to make it difficult for the mail delivery program to play
      with the sendmail config. To achieve this, we will run "mail delivery"
      in a chroot area UNDER the chroot area where sendmail itself is running.
      To do this, this program (localdeliver) must be setuid root in order to
      do the 2nd chroot successfully...
      
   c) localdeliver should be statically linked. No shared libraries please,
      as this may introduce extra vulnerabilities in the chroot area.

   d) Many dangerous parameters are pre-set in this source code.

       - the directory of the chroot area UNDER the sendmail chroot area
         is preset to "/local_delivery"
       - users who may receive local mail must be in group "system"
         (at least from the point of view of the sendmail chroot area and of
         the local_delivery chroot area)
       - the name of the final mail delivery program is preset to
         /bin/mail.local (from the point of view of the local_delivery chroot
         area)

   localdeliver performs the following actions in order:
   -----------------------------------------------------
   0. STARTS WITH ROOT PRIVILEGE, SINCE IT IS INVOKED SETUID
   1. does a basic scan of the command-line parameters, to extract the
      destination user name
   2. calculates the UNIX group id for group "system", the UNIX user id for
      the destination user name. It then chroots to "/local_delivery",
      and then drops privilege to user=dest username, group=system
   3. Rebuilts the original command-line parameters, and executes
      /bin/mail.local with those parameters
   4. By default, localdeliver does not allow mail delivery to the
      root account, since that would force it to run as root.
      This forces mail to root to be aliased to other accounts
      (local or remote).
*/

#include <syslog.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "drop_priv.h"

#ifdef USE_SYSLOG
#ifdef USE_PRTF
   I need either USE_SYSLOG or USE_PRTF defined, not both
#else
#endif
#else
#ifdef USE_PRTF
#else
   I need either USE_SYSLOG or USE_PRTF defined
#endif
#endif

char *called_prog_name = "/bin/mail.local";

void syntax( prgname )
char *prgname;
{
#ifdef USE_SYSLOG
      syslog( LOG_ERR,
         "Syntax: %s takes the same syntax as mail.local...\n", prgname );
#else
      fprintf( stderr,
         "Syntax: %s takes the same syntax as mail.local...\n", prgname );
#endif
      exit(1);
}

int main( argc, argv )
int argc;
char **argv;
{
   char *my_username = NULL;
   char *my_grpname  = NULL;
   char *my_dirname  = NULL;
   char *tmp_ptr     = NULL;

   int opt_rc;

   while ((opt_rc = getopt( argc, argv, "df:r:")) != EOF)
   {
      switch (opt_rc)
      {
         case 'd': break;
         case 'f':
         case 'r': tmp_ptr = optarg;
                   break;
         default : syntax( argv[0] );
                   break;
      }
   }

   if ((argv[optind]) == NULL) syntax( argv[0] );

   my_username  = argv[optind];

#ifndef ALLOW_ROOT_MAIL
   if (strcmp( my_username, "root" ) == 0)
   {
#ifdef USE_SYSLOG
      syslog( LOG_ERR,
         "Syntax: %s does not support mail to root account...\n", argv[0] );
#else
      fprintf( stderr,
         "Syntax: %s does not support mail to root account...\n", argv[0] );
#endif
      exit(1);
   }
#endif

   if (!drop_priv( my_username, "system", "/local_delivery" ))
   {
#ifdef USE_SYSLOG
      syslog( LOG_ERR, "Cannot drop privileges\n" );
#else
      fprintf( stderr, "Cannot drop privileges\n" );
#endif
      return 1;
   }

#ifdef 0
#ifdef USE_SYSLOG
      syslog( LOG_INFO, "After drop_priv" );
#else
      fprintf( stderr, "After drop_priv\n" );
#endif
#endif

   {
      char **arg_buf = NULL;
      int    our_idx = 0;

      arg_buf = (char **) malloc( sizeof(char *) * (argc + 1) );

      if (arg_buf == NULL)
      {
#ifdef USE_SYSLOG
         syslog( LOG_ERR, "Cannot allocate parameter area\n" );
#else
         fprintf( stderr, "Cannot allocate parameter area\n" );
#endif
         return 1;
      }

      arg_buf[ 0 ] = called_prog_name;

      for (our_idx = 1; our_idx < argc; our_idx++)
         arg_buf[ our_idx ] = argv[ our_idx ];

      arg_buf[ argc ] = NULL;

      (void) execv( called_prog_name, arg_buf );
   }
   return 1;
}
