/****************************** drop_priv.c ****************************/

/* Used if we need to run something with reduced privileges.
   Routine drop_priv should (successfully...) allow a process to change
   (in order)
              its root directory
              its group id
              its user id */

/******************************************************************************
 * Copyright (c) 1995,1996 Donated to the public domain                       *
 *                                                                            *
 * Original author and maintainer: Marc Chatel   (mchatel@dial.oleane.com)    *
 * Last known maintainer:          Marc Chatel   (mchatel@dial.oleane.com)    *
 *                                                                            *
 * This file was created as part of the S4 (Secure System Setup Script) kit.  *
 * Permission is granted to any person or entity to do any of the following:  *
 *       a) use this file alone or in some other software                     *
 *       b) modify this file or include parts of this file in other files     *
 *       c) re-distribute this file AS IS or modified, for non-commercial     *
 *          or commercial purposes, alone or as part of some software package *
 *                                                                            *
 * No warranties of any kind, express or implied, on the functionality and    *
 * safety of the contents of this file. Use at your own risk!                 *
 *                                                                            *
 * If you do useful changes to this file (bug fixes, portability fixes,       *
 * enhancements), you should TRY to contact the current maintainer, who may   *
 * be maintaining a "latest greatest" version of the file. You do not HAVE    *
 * TO, but you should TRY. Promote software reuse! It helps everybody,        *
 * including you!                                                             *
 ******************************************************************************/

#include <syslog.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>

#ifdef USE_SYSLOG
#ifdef USE_PRTF
   I need either USE_SYSLOG or USE_PRTF defined, not both
#else
#endif
#else
#ifdef USE_PRTF
#else
#define USE_SYSLOG
#endif
#endif

#define MAX_USERNAME_LEN    16
#define MAX_GRPNAME_LEN     16
#define MAX_DIRNAME_LEN    512
#define ERRBUF_LEN        1024

#define DEFAULT_USERNAME     "nobody"
#define DEFAULT_USERID       50000
#define DEFAULT_GRPNAME      "nogrp"
#define DEFAULT_GRPID        50000
#define DEFAULT_DIRNAME      "/tmp"

char  errbuf[ERRBUF_LEN];

void droppriv_errmsg()
{
#ifdef USE_SYSLOG
   syslog( LOG_ERR, errbuf );
#else
   fprintf( stderr, "%s\n", errbuf );
#endif

   errbuf[0] = '\0';
}

/************************** drop_priv() *******************************/

/* Reduce our privileges depending on specification.
   If username != NULL, find the equivalent userid and set our execution
      privileges to this userid. If we cannot find the username, use "nobody".
      If we cannot find "nobody", use 50000.
   If groupname != NULL, find the equivalent groupid and set our execution
      privileges to this groupid. If we cannot find the groupname, use 50000
      as a groupid.
   If dirname != NULL, change our root directory and our current directory to
      this directory name. If that fails, use /tmp as the new root.
   Of course, the actual order of operations is:
      get username->userid and groupname->groupid mapping
      change root directory and current directory
      change group id
      change user id
   Returns 1 if all OK, 0 if ANY problem. If problems, the routine will still
      attempt to reduce privilege to some level below root. */


int drop_priv( username, groupname, dirname )
char *username;
char *groupname;
char *dirname;
{
   uid_t          new_uid     = 0;
   uid_t          before_uid  = 0;
   gid_t          new_gid     = 0;
   int            rc          = 1; /* Successful by default */
   struct passwd *pwdfile_ptr = NULL;
   struct group  *grpfile_ptr = NULL;
   char           default_username[MAX_USERNAME_LEN+1];
   char           default_grpname[MAX_GRPNAME_LEN+1];
   char           default_dirname[MAX_DIRNAME_LEN+1];

   strcpy( default_username, DEFAULT_USERNAME );
   strcpy( default_grpname,  DEFAULT_GRPNAME  );
   strcpy( default_dirname,  DEFAULT_DIRNAME  );

   before_uid = geteuid();

   if (before_uid != 0)
   {
      strcpy( errbuf, 
              "BADCONFIG Need to begin as root in order to drop privileges");
      droppriv_errmsg();
      return 0;
   }

   if (username == NULL)
   {
      strcpy( errbuf, "BADCONFIG username parameter is NULL, using default" );
      droppriv_errmsg();
      username = &(default_username[0]);
      rc = 0;
   }

   if (strlen(username) > MAX_USERNAME_LEN)
   {
      strcpy( errbuf, "BADCONFIG username parameter too long, using default" );
      droppriv_errmsg();
      username = &(default_username[0]);
      rc = 0;
   }

   pwdfile_ptr = getpwnam( username );

   if (pwdfile_ptr != NULL)   new_uid = pwdfile_ptr->pw_uid;
   else
   {
      sprintf( errbuf, "BADCONFIG Cannot find username \"%s\", using default",
                       username );
      droppriv_errmsg();
      rc = 0;

      pwdfile_ptr = getpwnam( DEFAULT_USERNAME );

      if (pwdfile_ptr != NULL)  new_uid = pwdfile_ptr->pw_uid;
      else
      {
         sprintf( errbuf, "BADCONFIG Cannot find \"%s\", using userid %d",
                          DEFAULT_USERNAME, DEFAULT_USERID );
         droppriv_errmsg();
         new_uid = DEFAULT_USERID;
      }
   }

   pwdfile_ptr = NULL;
   endpwent();

   if (groupname == NULL)
   {
      strcpy( errbuf, "BADCONFIG groupname parameter is NULL, using default" );
      droppriv_errmsg();
      groupname = &(default_grpname[0]);
      rc = 0;
   }

   if (strlen(groupname) > MAX_GRPNAME_LEN)
   {
      strcpy( errbuf, "BADCONFIG groupname parameter too long, using default" );
      droppriv_errmsg();
      groupname = &(default_grpname[0]);
      rc = 0;
   }

   grpfile_ptr = getgrnam( groupname );

   if (grpfile_ptr != NULL)   new_gid = grpfile_ptr->gr_gid;
   else
   {
      sprintf( errbuf, "BADCONFIG Cannot find groupname \"%s\", using default",
                       groupname );
      droppriv_errmsg();
      rc = 0;

      grpfile_ptr = getgrnam( DEFAULT_GRPNAME );

      if (grpfile_ptr != NULL)  new_gid = grpfile_ptr->gr_gid;
      else
      {
         sprintf( errbuf, "BADCONFIG Cannot find \"%s\", using groupid %d",
                          DEFAULT_GRPNAME, DEFAULT_GRPID );
         droppriv_errmsg();
         new_gid = DEFAULT_GRPID;
      }
   }  

   if (initgroups( username, new_gid ) != 0)
   {
      sprintf( errbuf,
       "BADCONFIG initgroups failed on username \"%s\" and group %ld",
       username, (long) new_gid );

      droppriv_errmsg();
      rc = 0;
   }

   grpfile_ptr = NULL;
   endgrent();

   if (dirname != NULL)
   {
      if (strlen(dirname) > MAX_DIRNAME_LEN)
      {
         sprintf( errbuf, "BADCONFIG chroot dirname too long, using default" );
         droppriv_errmsg();
         rc = 0;

         dirname = &(default_dirname[0]);
      }

      if (chroot(dirname) != 0)
      {
         sprintf( errbuf, "BADCONFIG Cannot chroot to \"%s\", using default",
                          dirname );
         droppriv_errmsg();
         rc = 0;

         chroot( default_dirname );
      }

      /* Seems like the old filesystem is still accessible through
         the current directory ".". Change current directory to the
         new root directory. */

      if (chdir("/") != 0)
      {
         sprintf( errbuf,
                  "BUG Cannot change current directory to the new root" );
         droppriv_errmsg();
         rc = 0;
      }
   }

   if (setgid( new_gid ) != 0)
   {
      sprintf( errbuf, "BUG Failed to change group id to %d", new_gid );
      droppriv_errmsg();
      rc = 0;
   }

   if (setuid( new_uid ) != 0)
   {
      sprintf( errbuf, "BUG Failed to change user id to %d", new_uid );
      droppriv_errmsg();
      rc = 0;
   }

   return rc;
} /* drop_priv() */

/**************************** drop_priv_option() ***************************/

int drop_priv_option( my_option )
char *my_option;
{
   char *tmp_ptr     = NULL;
   char *my_username = NULL;
   char *my_grpname  = NULL;
   char *my_dir      = NULL;

   if (my_option == NULL)                              return 0;

   /* my_option begins with username. Find the beginning of group name */

   tmp_ptr = strchr( my_option, ':' );
   if (tmp_ptr == NULL)       return 0;  /* must find a colon */
   if (tmp_ptr == my_option)  return 0;  /* must have username before colon */

   *(tmp_ptr++) = '\0';      /* separate the username and groupname */
   my_username = my_option;

   /* remainder starts with groupname. 
      Find the beginning of the (optional) chroot directory */

   my_grpname = tmp_ptr;
   tmp_ptr = strchr( my_grpname, ':' );
   if (tmp_ptr == my_grpname)  return 0; /* must have groupname before colon */

   if (tmp_ptr != NULL)
   {
      /* we have a second colon, so we must have a chroot directory */

      *(tmp_ptr++) = '\0'; /* separate the groupname and chroot dir */

      my_dir = tmp_ptr;
      if ((*tmp_ptr) == '\0')  return 0; /* null chroot directory is not OK */
   }

   return drop_priv( my_username, my_grpname, my_dir );
} /* drop_priv_option() */

#ifdef MARC_DEBUG
/***************************** main() ****************************/

int main()
{
   int rc, fd;
   umask(0);
   openlog( "drop", LOG_PID, LOG_USER );
   syslog( LOG_DEBUG, "Starting up" );
   rc = drop_priv( "marc", "users", "/tmp" );
   syslog( LOG_DEBUG, "Creating file /test1" );
   fd = creat( "/test1", 0 );
   rc = drop_priv( "root", "root", "/usr" );
} /* main() */
#endif
