/********************* privportd.c **********************/

/******************************************************************************
 * Copyright (c) 1995,1996 Donated to the public domain                       *
 *                                                                            *
 * Original author and maintainer: Marc Chatel   (mchatel@dial.oleane.com)    *
 * Last known maintainer:          Marc Chatel   (mchatel@dial.oleane.com)    *
 *                                                                            *
 * This file was created as part of the S4 (Secure System Setup Script) kit.  *
 * Permission is granted to any person or entity to do any of the following:  *
 *       a) use this file alone or in some other software                     *
 *       b) modify this file or include parts of this file in other files     *
 *       c) re-distribute this file AS IS or modified, for non-commercial     *
 *          or commercial purposes, alone or as part of some software package *
 *                                                                            *
 * No warranties of any kind, express or implied, on the functionality and    *
 * safety of the contents of this file. Use at your own risk!                 *
 *                                                                            *
 * If you do useful changes to this file (bug fixes, portability fixes,       *
 * enhancements), you should TRY to contact the current maintainer, who may   *
 * be maintaining a "latest greatest" version of the file. You do not HAVE    *
 * TO, but you should TRY. Promote software reuse! It helps everybody,        *
 * including you!                                                             *
 ******************************************************************************/

/* This program is a daemon that runs as root. BEWARE when modifying the code!
   The purpose of this daemon is as follows:

     a) read the file /etc/privport.conf to get certain configuration details
        (basically the list of ports and addresses that the daemon will be
        allowed to give to clients)

     b) open a UNIX domain STREAM socket called /dev/privport
        and listen for requests

     c) when somebody connects to this socket, fork a child to handle the
        request. The request will be to get a socket of a specific type,
        to do a bind on a specific port (UDP or TCP), and to pass back the
        open socket to the requesting process.

        A request should be a single line followed by a line feed.

        The line should contain three values separated by spaces:

           a character which will be "U" for UDP or "T" for TCP
           a port number in ASCII
           an IP address in ASCII (if you wish to bind against INADDR_ANY,
              just pass the value of INADDR_ANY in ASCII form)
   
   I place this code in the public domain, free to everybody.

               Marc Chatel, 1995 */

/* Daemon startup code taken almost verbatim from the
   "Unix Network Programming" book,
   Author W. Richard Stevens, Prentice Hall, ISBN 0-13-949876-1

   Anybody who reads this code and does not have this book yet
   should run and BUY that book. It is almost a prerequisite for
   this type of code.

   This is a skeleton of how a UNIX daemon process is supposed to be
   coded. This example is on pages 72-85 of the book.

   Typed in by Marc Chatel... */

   
#include   <standards.h>
#include   <sys/types.h>
#include   <sys/resource.h>
#include   <wait.h>
#include   <stdio.h>
#include   <sys/wait.h>
#include   <signal.h>
#include   <sys/param.h>
#include   <errno.h>
#include   <syslog.h>
#include   <sys/socket.h>
#include   <string.h>
#include   <sys/un.h>
#include   <unistd.h>
#include   <fcntl.h>
#include   <netinet/in.h>
#include   <arpa/inet.h>
#include   <errno.h>
#include   <sys/uio.h>
#include   <sys/stat.h>

#ifdef SIGTSTP   /* True if BSD system */
#include  <sys/file.h>
#include  <sys/ioctl.h>
#endif


#define MAX_ALLOWED_SOCKETS     20
#define PRIVPORT_SOCKET         "/dev/privport"
#define MAX_BUF_LEN             200


struct  socket_struct
{
   int            socktype; /* SOCK_STREAM or SOCK_DGRAM */
   u_short        port;     /* UDP or TCP port in network byte order */
   struct in_addr inaddr;   /* IP address in network byte order */
};

int num_allowed_sockets = 0;

struct socket_struct allowed_sockets[MAX_ALLOWED_SOCKETS];

int debug = 0;

/*********************** sig_child() *****************************/

int sig_child()
{
   pid_t   pid;
   int     status;

   if (debug) syslog( LOG_DEBUG, "sig_child() called" );

   while (1)
   {
      pid = waitpid( (pid_t) -1, &status, WNOHANG );
      if (debug) syslog( LOG_DEBUG, "sig_child: pid = %ld", (long) pid );
      if (pid > 0) break;
   }

   return 0;
} /* sig_child() */


/***************************** daemon_start() *******************************/

/* Detach a daemon process from login session context */

void daemon_start( ignsigcld )
int  ignsigcld;   /* nonzero -> handle SIGCLDs so zombies don't clog */
{
   register int   childpid, fd;

   /* If we were started by init (process 1) from the /etc/inittab file
      there is no need to detach. This test is unreliable due to an
      unavoidable ambiguity if the process is started by some other process
      and orphaned (i.e. if the parent process terminates before we are
      started) */

   if (getppid() == 1)  goto out;

   /* Ignore the terminal stop signals (BSD) */

#ifdef SIGTTOU
   signal( SIGTTOU, SIG_IGN );
#endif
#ifdef SIGTTIN
   signal( SIGTTIN, SIG_IGN );
#endif
#ifdef SIGTSTP
   signal( SIGTSTP, SIG_IGN );
#endif

   /* If we were not started in the background, fork and let the parent exit.
      This also guarantees the first child is not a process group leader. */

   if ((childpid = fork()) < 0)
   {
      syslog( LOG_ERR, "can't form first child: %m" );
      exit(1);
   }

   if (childpid > 0)
   {
      if (debug) syslog( LOG_DEBUG, "Parent exiting" );
      exit(0);  /* parent; */
   }

   if (debug) syslog (LOG_DEBUG, "We are the first child" );

   /* First child process. Disassociate from controlling terminal and
      process group. Ensure the process can't reacquire a new controlling
      terminal. */

#ifdef SIGTSTP   /* BSD */
   if (setpgrp( 0,getpid() ) == -1)
   {
      syslog( LOG_ERR, "can't change process group: %m" );
      exit(1);
   }

   if ((fd = open("/dev/tty",O_RDWR)) >= 0)
   {
      ioctl(fd, TIOCNOTTY, (char *)NULL); /* lose controlling tty */
      close(fd);
   }

   if (debug) syslog( LOG_DEBUG, "BSD style tty disconnect OK" );
#else   /* System V */
   if (setpgrp() == -1)
   {
      syslog( LOG_ERR, "can't change process group: %m" );
      exit(1);
   }

   if (debug) syslog( LOG_DEBUG, "changed process group with setpgrp() OK" );

   signal(SIGHUP, SIG_IGN); /* immune from pgrp leader death */

   if ((childpid = fork()) < 0)
   {
      syslog( LOG_ERR, "can't fork second child: %m" );
      exit(1);
   }
   else
      if (childpid > 0)
      {
         if (debug) syslog( LOG_DEBUG, "First child exiting" );
         exit(0);   /* first child */
      }

   /* second child */

   if (debug) syslog( LOG_DEBUG, "Second child started up" );
#endif

   out:
   /* Close common open file descriptors */
   close(0); close(1); close(2);
  
   /* This is effective too, but breaks SYSLOG traffic */ 
   /***  for (fd = 0; fd < NOFILE; fd++)   close(fd); ***/

   errno = 0;  /* probably got set to EBADF from a close */

   if (debug) syslog( LOG_DEBUG, "Closed file descriptors" );

   /* Move the current directory to root, to make sure we are not
      on a mounted filesystem */

   chdir("/");

   /* Clear any inherited file mode creation mask. */

   umask(0);

   /* See if the caller isn't interested in the exit status of its
      children, and does not want to have them become zombies and clog up
      the system. With System V all we need to do is ignore the signal.
      With BSD, however, we have to catch each signal and execute the
      wait3() system call. */

   if (ignsigcld)
   {
      struct sigaction  new_sigaction, old_sigaction;

      if (sigaction( SIGCHLD, (struct sigaction *) NULL,
                     &old_sigaction ) != 0)
      {
         syslog( LOG_ERR, "First SIGCHLD sigaction failed" );
         exit(1);
      }

      new_sigaction = old_sigaction;

      new_sigaction.sa_flags |= SA_NOCLDWAIT;
      new_sigaction.sa_handler = SIG_IGN;

      if (sigaction( SIGCHLD, &new_sigaction, &old_sigaction ) != 0)
      {
         syslog( LOG_ERR, "Second SIGCHLD sigaction failed" );
         exit(1);
      }

#ifdef 0
#ifdef SIGTSTP
      int sig_child();

      signal( SIGCLD, sig_child );   /* BSD */
#else
      signal( SIGCLD, SIG_IGN );   /* System V */
#endif
#endif
   }
} /* daemon_start() */


/****************************** parse_line() ********************************/

/* parse_line: Consider str to be an input line. If it conforms to the format
               [U|T]  port_number  ip_address, updates the parameters
               *socktype_ptr, *port_ptr and *inaddr_ptr 

   Returns 1 if OK, 0 if things went wrong */

int parse_line( str, socktype_ptr, port_ptr, inaddr_ptr )
char           *str;
int            *socktype_ptr;
u_short        *port_ptr;
struct in_addr *inaddr_ptr;
{
   if (str == NULL)
   {
      syslog( LOG_ERR, "parse_line: str is NULL" );
      return 0;
   }

   if (socktype_ptr == NULL)
   {
      syslog( LOG_ERR, "parse_line: socktype_ptr is NULL" );
      return 0;
   }

   if (port_ptr == NULL)
   {
      syslog( LOG_ERR, "parse_line: port_ptr is NULL" );
      return 0;
   }

   if (inaddr_ptr == NULL)
   {
      syslog( LOG_ERR, "parse_line: inaddr_ptr is NULL" );
      return 0;
   }

   if (debug) syslog( LOG_DEBUG, "parse_line: input is \"%s\"", str );

   /* Skip spaces and tabs */
   while ((*str == ' ') || (*str == '\t'))   str++;

   switch (*str)
   {
      case 'U':
      case 'u': *socktype_ptr = SOCK_DGRAM; break;
      case 'T':
      case 't': *socktype_ptr = SOCK_STREAM; break;
      default:  syslog( LOG_ERR, "parse_line: expected U or T" );
                return 0;
                break;
   }
   str++; /* skip the socket type */

   if ((*str != ' ') && (*str != '\t'))
   {
      syslog( LOG_ERR,"parse_line: expected space after socket type" );
      return 0;
   }
   str++;

   /* Skip additional spaces and tabs */
   while ((*str == ' ') || (*str == '\t'))  str++;

   /* Read port number */
   {
      u_short tmp_val = 0;

      if ((*str < '0') || (*str > '9'))
      {
         syslog( LOG_ERR,
                 "parse_line: port number must start with digit" );
         return 0;
      }

      while ((*str >= '0') && (*str <= '9'))
      {
         tmp_val = tmp_val * 10;
         tmp_val = tmp_val + ((*str) - '0');
         str++;
         if (tmp_val > 9999) break;   /* do not allow overflow */
      }

      if ((*str != ' ') && (*str != '\t'))
      {
         syslog( LOG_ERR, "parse_line: invalid port number" );
         return 0;
      }

      *port_ptr = htons( tmp_val );
   }

   /* Skip additional spaces and tabs */
   while ((*str == ' ') || (*str == '\t'))  str++;

   if ((*str < '0') || (*str > '9'))
   {
      syslog( LOG_ERR, "parse_line: IP address must start with digit" );
      return 0;
   }

   {
      char *begin_ip_addr;

      begin_ip_addr = str;

      while (((*str >= '0') && (*str <= '9')) || (*str == '.'))
         str++;

      if ((*str != ' ') && (*str != '\t') && (*str != '\n') &&
          (*str != '\0'))
      {
         syslog( LOG_ERR, "parse_line: IP address looks invalid" );
         return 0;
      }

      /* inet_addr is supposed to return a value in network byte
         order, so we have nothing to do... */

      inaddr_ptr->s_addr = inet_addr( begin_ip_addr );

      if (inaddr_ptr->s_addr == INADDR_NONE)
      {
         syslog( LOG_ERR, "parse_line: inet_addr failed" );
         return 0;
      }
   }

   if (debug) syslog( LOG_DEBUG, "parse_line returns OK" );
   return 1;
} /* parse_line() */


/*************************** read_config() ******************************/

/* read_config: Reads in the file /etc/privport.conf and loads the table
                allowed_sockets. Returns 1 if OK, 0 if failed. */

int read_config()
{

   FILE *my_file = NULL;
   char  my_buf[MAX_BUF_LEN];
   int   our_rc = 1;

   my_file = fopen( "/etc/privport.conf", "r" );

   if (my_file == NULL)
   {
      syslog( LOG_ERR, "read_config: cannot open /etc/privport.conf: %m" );
      return 0;
   }

   while (fgets( my_buf, MAX_BUF_LEN, my_file ) != NULL)
   {
      int             my_len;
      int             tmp_sock_type;
      u_short         tmp_port_num;
      struct in_addr  tmp_ip_addr;

      my_buf[MAX_BUF_LEN - 1] = '\0';
      my_len = strlen( my_buf );

      if (my_len <= 0)
      {
         syslog( LOG_ERR, "read_config: line length <= 0" );
         our_rc = 0; break;
      }

      if (my_buf[my_len - 1] != '\n')
      {
         syslog( LOG_ERR, "read_config: line too long" );
         our_rc = 0; break;
      }

      my_buf[my_len - 1] = '\0'; /* remove carriage return */

      if (my_buf[0] == '#')
      {
         /* Comment line, ignore it */
      }
      else
      {
         if (!parse_line(my_buf, &tmp_sock_type, &tmp_port_num, &tmp_ip_addr))
         {
            syslog( LOG_ERR, "read_config: parse_line failed" );
            our_rc = 0; break;
         }

         if (num_allowed_sockets >= MAX_ALLOWED_SOCKETS)
         {
            syslog( LOG_ERR, "read_config: more than %d authorized sockets",
                    MAX_ALLOWED_SOCKETS );
            our_rc = 0; break;
         }

         allowed_sockets[num_allowed_sockets].socktype = tmp_sock_type;
         allowed_sockets[num_allowed_sockets].port     = tmp_port_num;
         allowed_sockets[num_allowed_sockets].inaddr   = tmp_ip_addr;

         num_allowed_sockets++;
      }
   }

   fclose( my_file );

   if (debug) syslog( LOG_DEBUG, "read_config: finishing, rc=%d", our_rc );
   return our_rc;
} /* read_config() */


/************* send_socket_and_close_it() ************************/

int send_socket_and_close_it( comm_socket, send_fd )
int comm_socket, send_fd;
{
   struct iovec    iov[1];
   struct msghdr   msg;
   char            tmp_buf[2];
   int             send_rc;
   int             our_rc = 0;

   if (send_fd < 0)    tmp_buf[0] = 'B';  /* passing a bad socket */
   else                tmp_buf[0] = 'G';  /* passing a good socket */

   tmp_buf[1] = '\0';

   iov[0].iov_base = (char *) (&tmp_buf[0]); /* send one byte */
   iov[0].iov_len  = 1;

   msg.msg_iov      = iov;
   msg.msg_iovlen   = 1;
   msg.msg_name     = (caddr_t) 0;

   /* If good, Pass the address of the GRANT socket.
      Also indicate we are passing ONE descriptor. */

#ifdef _SOCKADDR_LEN
   if (send_fd < 0)
   {
      msg.msg_control     = (caddr_t) NULL;
      msg.msg_controllen  = 0;
   }
   else
   {
      msg.msg_control    = (caddr_t) &send_fd;
      msg.msg_controllen  = sizeof(send_fd);
   }
#else
   if (send_fd < 0)
   {
      msg.msg_accrights     = (caddr_t) NULL;
      msg.msg_accrightslen  = 0;
   }
   else
   {
      msg.msg_accrights     = (caddr_t) &send_fd;
      msg.msg_accrightslen  = sizeof(send_fd);
   }
#endif

   send_rc = sendmsg( comm_socket, &msg, 0 );

   if (debug)
      syslog( LOG_DEBUG, "called sendmsg() with fd=%d, passing fd=%d, rc=%d",
                         comm_socket, send_fd, send_rc );

   if (send_rc < 0)
   {
      syslog( LOG_ERR, "send_socket_and_close_it: sendmsg() failed: %m" );
   }
   else
   {
      if (debug)
         syslog( LOG_DEBUG, "send_socket_and_close_it: sendmsg() OK") ;

      our_rc = 1;
   }

   if (send_fd >= 0)
   {
      close( send_fd );
      if (debug)
      {
         syslog( LOG_DEBUG, "closed send_fd" );
         /* sleep(3); */
      }
   }

   return our_rc;
}  /* send_socket_and_close_it() */


/************************** handle_requests() ************************/

void handle_requests( comm_socket )
int comm_socket;
{
   char            my_buf[MAX_BUF_LEN];
   int             cur_len = 0;
   int             req_socktype = 0;
   char            text_socktype[5];
   char            decision[20];
   u_short         req_port = 0;
   struct in_addr  req_ip_addr;
   int             srch_idx, found_it;

   while (1)
   {
      int read_rc;

      if (debug) syslog( LOG_DEBUG, "handle_requests: top of while(1)" );

      cur_len = 0;
      my_buf[0] = '\0';

      if (comm_socket < 0)
      {
         syslog( LOG_ERR, "handle_requests: comm_socket is negative" );
         return;
      }

      read_rc = read( comm_socket, my_buf + cur_len, 1 );

      while (read_rc == 1)
      {
         cur_len++;
         my_buf[cur_len] = '\0';

         if (my_buf[ cur_len - 1 ] == '\n')    break;

         if ((cur_len + 1) >= MAX_BUF_LEN)
         {
            syslog( LOG_ERR, "handle_requests: line is too long" );
            return;
         }

         read_rc = read( comm_socket, my_buf + cur_len, 1 );
      } /* while read() == 1 */

      if (read_rc != 1)
      {
         syslog( LOG_INFO,
                 "No more client data or error, closing client stream" );
         return;
      }

      /* we got a complete line from the requestor */

      if (!parse_line( my_buf, &req_socktype, &req_port, &req_ip_addr ))
      {
         syslog( LOG_ERR, "handle_requests: parse_line() failed" );
         return;
      }

      if (req_socktype == SOCK_DGRAM)  strcpy( text_socktype, "UDP" );
      else                             strcpy( text_socktype, "TCP" );

      /* Check if the request is in our list of allowed sockets */

      found_it = 0;

      for (srch_idx = 0; srch_idx < num_allowed_sockets; srch_idx++ )
      {
         if ((req_socktype       == allowed_sockets[srch_idx].socktype) &&
             (req_port           == allowed_sockets[srch_idx].port)     &&
             (req_ip_addr.s_addr == allowed_sockets[srch_idx].inaddr.s_addr))
         {
            found_it = 1;
            break;
         }
      } /* for srch_idx = 0 to num_allowed_sockets - 1 */

      if (found_it)   strcpy( decision, "Accepted" );
      else            strcpy( decision, "REJECTED" );

      syslog( LOG_INFO, "%s request for %s port %d on IP address %s",
                        decision, text_socktype,
                        ntohs( req_port ), inet_ntoa( req_ip_addr  ) );

/*    if (debug) sleep(3); */

      if (!found_it)
      {
         if (!send_socket_and_close_it( comm_socket, -1 ))
         {
            syslog( LOG_ERR, "handle_requests: send_socket() failed" );
         }
         return;
      }
      else
      {
         /* We accept the request. Do the work. */

         int                 grant_socket = -1;
         int                 tmp_opt_buf;
         struct sockaddr_in  grant_sockaddr;

         grant_socket = socket( AF_INET, req_socktype, 0 );

         if (grant_socket < 0)
         {
            syslog( LOG_ERR, "handle_requests: socket() for grant failed: %m" );
            return;
         }

         if (debug)
         {
            syslog( LOG_DEBUG,
                    "handle_requests: socket() OK, fd=%d", grant_socket);
            /* sleep(2); */
         }

         if (setsockopt( grant_socket, SOL_SOCKET, SO_REUSEADDR,
                         (char *) &tmp_opt_buf, sizeof( tmp_opt_buf ) ) < 0)
         {
            syslog( LOG_ERR, "handle_requests: setsockopt() failed: %m" );
            return;
         }

         if (debug)
         {
            syslog( LOG_DEBUG, "handle_requests: setsockopt OK" );
            /* sleep(2); */
         }

         grant_sockaddr.sin_family = AF_INET;
         grant_sockaddr.sin_port   = req_port;
         grant_sockaddr.sin_addr   = req_ip_addr;

         if (bind( grant_socket, (struct sockaddr *) &grant_sockaddr,
                   sizeof( grant_sockaddr ) ) < 0 )
         {
            syslog( LOG_ERR, "handle_requests: bind() for grant failed: %m" );

            sleep(3);

            if (bind( grant_socket, (struct sockaddr *) &grant_sockaddr,
                      sizeof( grant_sockaddr ) ) < 0 )
            {
               syslog( LOG_ERR, "handle_requests: bind() second attempt: %m" );
               return;
            }
         }

         if (debug)
         {
            syslog( LOG_DEBUG, "handle_requests: bind OK" );
            /* sleep(4); */
         }

         if (!send_socket_and_close_it( comm_socket, grant_socket ))
         {
            syslog( LOG_ERR, "handle_requests: send_socket() failed" );
         }
      } /* if found_it */
   } /* while (1) */
} /* handle_requests() */


/******************************** main() *********************************/

int main( argc, argv )
int argc;
char **argv;
{
   int                 privport_socket = -1;
   struct sockaddr_un  privport_sockaddr;
   struct sockaddr_un  client_sockaddr;
   int                 privport_fn_len;
   int                 new_socket = -1;
   int                 junk_val;

   openlog( basename(argv[0]),
            LOG_PID | LOG_NDELAY | LOG_NOWAIT,
            LOG_DAEMON );

   if (debug) syslog( LOG_DEBUG, "Starting up!" );

   if (!read_config())
   {
      syslog( LOG_ERR, "main: read_config() failed" );
      return 1;
   }

   if (debug) syslog( LOG_DEBUG, "Config read OK" );

   daemon_start(1); /* make sure zombies are handled */

   if (debug) syslog( LOG_DEBUG, "Demonized OK" );

   {
      struct stat tmp_stat_buf;

      if (stat( PRIVPORT_SOCKET, &tmp_stat_buf ) == 0)
      {
         if (debug) syslog( LOG_DEBUG, "Found old privport socket, deleted" );
         unlink( PRIVPORT_SOCKET );
      }
      else
      {
         if (debug)
            syslog( LOG_DEBUG, "Privport socket did not exit, creating" );
      }
   }

   privport_fn_len = strlen( PRIVPORT_SOCKET );
   privport_sockaddr.sun_family = AF_UNIX;
   strcpy( privport_sockaddr.sun_path, PRIVPORT_SOCKET );

#ifdef _SOCKADDR_LEN
   privport_sockaddr.sun_len = privport_fn_len + 1;
#endif

   privport_socket = socket( AF_UNIX, SOCK_STREAM, 0 );

   if (privport_socket < 0)
   {
      syslog( LOG_ERR, "main: privport socket() failed: %m" );
      return 1;
   }

   if (debug) syslog( LOG_DEBUG, "privport socket() gotten OK, fd=%d",
                                  privport_socket );

   {
      int tmp_opt_buf;

      if (setsockopt( privport_socket, SOL_SOCKET, SO_REUSEADDR,
                    (char *) &tmp_opt_buf, sizeof( tmp_opt_buf ) ) < 0)
      {
         syslog( LOG_ERR, "main: privport setsockopt() failed: %m" );
         return 1;
      }
   }
   if (debug) syslog( LOG_DEBUG, "privport setsockopt() OK" );

   if (bind( privport_socket,
             (struct sockaddr *) &privport_sockaddr,
             sizeof( struct sockaddr_un )) != 0)
   {
      syslog( LOG_ERR, "main: privport bind() failed: %m" );
      return 1;
   }

   if (debug) syslog( LOG_DEBUG, "privport bind() OK" );

   if (listen( privport_socket, SOMAXCONN ) != 0)
   {
      syslog( LOG_ERR, "main: privport listen() failed: %m" );
      return 1;
   }

   if (debug) syslog( LOG_DEBUG, "privport listen() OK" );

   syslog( LOG_INFO, "Ready to accept requests" );

   junk_val = sizeof( struct sockaddr_un );

   while (new_socket = accept( privport_socket,
                               (struct sockaddr *) &client_sockaddr,
                               &junk_val ))
   {
      if (debug)
      {
         if (new_socket >= 0)
            syslog( LOG_DEBUG, "accept() returns %d", new_socket );
         else
            syslog( LOG_DEBUG, "accept() returns %d, errmsg=%m", new_socket );

         /* sleep(5); */
      }

      if (new_socket < 0)
      {
         if (errno != EINTR) break;
      }
      else
      {
         pid_t new_process;

         junk_val = sizeof( struct sockaddr_un );

         new_process = fork();

         switch (new_process)
         {
            case -1: syslog( LOG_ERR,
                             "main: fork() failed, refusing request: %m" );
                     close( new_socket );
                     break;
            case  0: /* we are the child */
                     if (debug)
                     {
                        /* sleep(4); */
                        syslog( LOG_DEBUG,
                           "main: child starting up to handle request" );
                     }
                     close( privport_socket );
                     handle_requests( new_socket );
                     close( new_socket );
                     return 0; /* exit */
                     break;
            default: /* we are the parent */
                     if (debug)
                     {
                        /* sleep(2); */
                        syslog( LOG_DEBUG,
                          "main: parent of the fork() for handle request" );
                     }
                     close( new_socket ); /* we do not need this */
                     break;
         }
      }
   }

   syslog( LOG_ERR, "main: privport accept() failed: %m" );
   return 1;
} /* main() */
